// This file is part of Desktop App Toolkit,
// a set of libraries for developing nice desktop applications.
//
// For license and copyright information please follow this link:
// https://github.com/desktop-app/legal/blob/master/LEGAL
//

#include <QtCore/QString>
#include <QtCore/QDateTime>

#include <rpl/rpl.h>

#include "base/basic_types.h"
