// This file is part of Desktop App Toolkit,
// a set of libraries for developing nice desktop applications.
//
// For license and copyright information please follow this link:
// https://github.com/desktop-app/legal/blob/master/LEGAL
//
#include "base/platform/linux/base_layout_switch_linux.h"

namespace base::Platform {

bool SwitchKeyboardLayoutToEnglish() {
	return false;
}

} // namespace base::Platform
